﻿#pragma once
#include <string>
#include <math.h>
#include "DataTyps.h"
#include "MathExt.h"
#include <QApplication>
#include <QClipboard>
#include <stdio.h>
#include <QDir>
#include <QFile>
#include <QTextStream>
using namespace std;
//! holds data grid information
typedef struct rect{double minlat, maxlat,minlon,maxlon;}rect;
//! holds data grid information with angles multiplied by 1200
typedef struct recti{int minlat, maxlat,minlon,maxlon;}recti;
//! holds weights for cubic interpolation
typedef struct CubicKernel3{int index[6]; double weights[6];}CubicKernel3;
//! UTM data container
typedef struct UTM{int x;int y; float h;}UTM;


//! holds digital elevation model extracted from SRTM3 data
/*!
  acts on requests concerning heightdata at point (lon,lat). hgt files of SRTM3
  data are used to generate height data grid, being interpolateted either
  lineary or bicubic. There exist fast serving functions acting on
  precalculated data grid labeled with "_t" (=templat) at the end of its name.
  height() and height_t() not only extract the height but can as well extract
  the gradient and the reliability of the heightdata with respect to its
  direction. This has to be specified by the mode-argument.
*/
class HeightData
{
public:
    void test_relax_grid();
	HeightData(void);
	~HeightData(void);
    //! returns height data
    /*! Depending on the selected mode the following heigt data is returned:\n
        mode 0: height\n
        mode 1: longitudinal descent [m /(°/1200)]\n
        mode 2: latitudinal descent [m /(°/1200)]\n
        mode 3: azimut in [rad] of the gradient\n
        mode 4: takes hc = double[3] with hc[1] = latitude part and
                hc[2] = longitude part to define direction, returns height and
                saves mistrust indec in hc[0]\n
        mistrust index = abs(rx*gy-ry*gx)*grr with rx,ry...road direction, gx
        and gy as the DEM gradient and grr the 2nd derivativ of DEM in road
        direction indicating how fast the gradient will change.\n
        This function uses bicubic convolution of 36 grid points. Internally
        for every request the related grid is created making this function
        slow. For fast purpose use height_t() instead.
    */
    double height(double lon,double lat,void* hc,short mode);
    double simple_height(double lon,double lat);
    //! specifies the folder where hgt-files of SRTM3 data are searched
    void set_SRTM3_folder(char* folder);
    //! specifies the folder where hgt-files of SRTM1 data are searched
    void set_SRTM1_folder(char* folder);
    //! specifies the folder where hgt-files of SRTM066 data are searched
    void set_SRTM066_folder(char* folder);
    //! specifies the folder where hgt-files of SRTM018 data are searched
    void set_SRTM018_folder(char* folder);
    //! specifies area of the precalculated data grid
	void set_rect(double minlat,double maxlat,double minlon,double maxlon);
    //! generates the data grid being used by t-function
	void reset_t(double minlat,double maxlat,double minlon,double maxlon);
    //! maximum height of the precalculated data grid
    int max_t();
    //! maximum height difference of the precalculated data grid
    int max_dt();
    //! returns height data operating at the precalculated data grid
    /*!
        base function of height. Operates on a precalculated data grid an takes
        the same requests as height().
    */
    double height_t(double lon,double lat,void* hc,short mode);
    //! with extracheck for coordinates in data range
    double height_t_save(double lon,double lat,void* hc,short mode);
    //! evaluates prominence of summits
    void paek_pro(double lon, double lat, qint64** c,float* Ahigher, int cc,
				  int ac, float d,float mperlon, float mperlat);
    //! returns heigth data grid h with size (width,height)
    /*! interpolating lineary the raw data. Data area has to be set with
     *  set_rect() command. Gradient out put dhdx and dhdy is optional, set
        NULL if not required. Constant y spacing is applied.
    */
	void interpolate_linear(int width,int height,double alpha,
							double** h, double** dhdx, double** dhdy);
    //! returns heigth data grid h with size (width,height)
    /*! interpolating lineary the raw data. Data area has to be set with
     *  set_rect() command. Gradient out put dhdx and dhdy is optional, set
        NULL if not required. Real variable y mercator spacing is applied.
    */
    void interpolate_linear_mercator(int width,int height,double alpha,
                            double** h, double** dhdx, double** dhdy);
    //! returns heigth data grid h with size (width,height)
    /*! interpolating bicubic the raw data. Data area has to be set with
        set_rect() command. Gradient out put dhdx and dhdy is optional, set
        NULL if not required.
    */
	void interpolate_bicubic_conv3(int width,int height,double alpha,
								   double** h, double** dhdx, double** dhdy);
    //! returns heigth data grid h with size (width,height)
    /*! interpolating bicubic the raw data. Data area has to be set with
        set_rect() command. Gradient out put dhdx and dhdy is optional, set
        NULL if not required. Constant y spacing is applied.
    */
    void interpolate_bicubic_conv3_mercator(int width,int height,double alpha,
                                   double** h, double** dhdx, double** dhdy);
    //! returns sorted raw data heights within the specified area
    /*! create grid has to be called before. Data arrays are allocated by the
        function. Uses quicksort algorithm to sort.  Real variable y mercator
        spacing is applied.
    */
    void get_sorted_grid_data(short** heights, int** lons, int** lats,
                              int** gradx, int** grady, long* count);
	recti get_GridRect();
    int correct_hgt(char* filename);
    char SRTM3folder[1000];             /*!< folder to contain files*/
    void DOM_to_SRTM(QString DOM_src_folder, QString SRTM_dest_folder,
                     int DOMresolution, int SRTMresolution);
    //! transformation of Saxony DEM data to SRTM data with arbitary resolution
    /*! reads in full DOM, recognizes suitable SRTM borders and creates
     *  corresponding SRTM-files. Resolution of SRTM file is to be specified,
     *  e.g. SRTM3=1200, SRTM1=3600. BTP supports SRTM018=20000 and SRTM066=5500
     *  for base map creating. Attention: To increase height resolution data
     *  format of hgt-files is NOT normal int16, but height is to be read as
     *  (int32+32676)/32
    */
    void UTM_to_latlon(int UTMcell, double x, double y, double *lat, double* lon);
    void latlon_to_UTM(int*UTMcell, double* x, double* y, double lat, double lon);
    double FootpointLatitude(double y);
    double ArcLengthOfMeridian (double phi);
private:
    int sx,sy,nReserve;                 /*!< dimensions of data grid*/
    rect DEMRect;                       /*!< coordinates of data grid*/
    recti GridRect;                     /*!< coordinates*1200 of data grid*/
    int gcount;                         /*!< grid count 1200 or 3601*/
    double gcountf;                     /*!< float version of gcount*/
    float h_scale;                      /*!< height scaling factor for SRTM018*/

    void create_grid();                 /*!< creates data grid*/
    void free_grid();                   /*!< deletes data grid*/
    float**  grid;                      /*!< data grid*/
    int lty(double lat);                /*!< bottom data grid index*/
    int ltx(double lon);                /*!< left data grid index*/
    double ltdy(double lat);            /*!< weight of bottom data grid point*/
    double ltdx(double lon);            /*!< weight of left data grid point*/

    //! quick sort used by get_sorted_grid_data()
	void sort_GridList(long s, long e);
    short *qsh; int *qslon,*qslat; int *qsy,*qsx;
    //! calculates weights of the bicubic convolution for the absolute height
    double Cubic_conv_kernel3(double x);
    //! calculates weights of the bicubic convolution for the gradient
    double Cubic_conv_kernel3_d(double x);
    //! executes the bicubic convolution with the kernels kx and ky
    double Bicubic_conv3(CubicKernel3 kx, CubicKernel3 ky);
    //! removes cos-like height artifacts on grid, use for SRTM1 bejond 80° lat
    void remove_wave_artifacts();
    //! relaxes grid bumps caused by disrectized height data
    void relax_grid_bumps();
    void relax_grid();
};

